#!/usr/bin/env python3
#
# Copyright (C) 2008-2020 Sébastien Helleu <flashcode@flashtux.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""
Documentation generator for WeeChat: build include files with:

- commands
- config options
- default aliases
- IRC colors
- infos
- infos hashtable
- infolists
- hdata
- completions
- URL options
- plugins priority.

Instructions to build config files yourself in WeeChat directories
(replace "path" with the path to the docgen.py script in WeeChat repository):

  weechat -t -r "/python load /path/docgen.py;/docgen;/quit"

There's one output file per language (where xx is language):

  /path/xx/includes/autogen.xx.adoc

This script requires Python 3.6+.
"""

SCRIPT_NAME = 'docgen'
SCRIPT_AUTHOR = 'Sébastien Helleu <flashcode@flashtux.org>'
SCRIPT_VERSION = '0.3'
SCRIPT_LICENSE = 'GPL3'
SCRIPT_DESC = 'Documentation generator for WeeChat'

SCRIPT_COMMAND = 'docgen'

IMPORT_OK = True

# pylint: disable=wrong-import-position
try:
    from collections import defaultdict
    from operator import itemgetter
    import gettext
    import hashlib
    import inspect
    import os
    import re
except ImportError as exc:
    print(f'Missing package(s) for {SCRIPT_NAME}: {exc}')
    IMPORT_OK = False

try:
    import weechat  # pylint: disable=import-error
except ImportError:
    print('This script must be run under WeeChat.')
    print('Get WeeChat now at: https://weechat.org/')
    IMPORT_OK = False

# list of locales for which we want to build doc files to include
LOCALE_LIST = (
    'de_DE',
    'en_US',
    'fr_FR',
    'it_IT',
    'ja_JP',
    'pl_PL',
)

# all commands/options/.. of following plugins will produce a file
# non-listed plugins will be ignored
# value: "c" = plugin may have many commands
#        "o" = write config options for plugin
# if plugin is listed without "c", that means plugin has only one command
# /name (where "name" is name of plugin)
# Note: we consider core is a plugin called "weechat"
PLUGIN_LIST = {
    'sec': 'o',
    'weechat': 'co',
    'alias': '',
    'buflist': 'co',
    'charset': 'o',
    'exec': 'o',
    'fifo': 'o',
    'fset': 'o',
    'irc': 'co',
    'logger': 'o',
    'relay': 'o',
    'script': 'o',
    'perl': 'o',
    'python': 'o',
    'ruby': 'o',
    'lua': 'o',
    'tcl': 'o',
    'guile': 'o',
    'javascript': 'o',
    'php': 'o',
    'spell': 'o',
    'trigger': 'o',
    'xfer': 'co',
}

# options to ignore
IGNORE_OPTIONS = (
    r'charset\.decode\..*',
    r'charset\.encode\..*',
    r'irc\.msgbuffer\..*',
    r'irc\.ctcp\..*',
    r'irc\.ignore\..*',
    r'irc\.server\..*',
    r'logger\.level\..*',
    r'logger\.mask\..*',
    r'relay\.port\..*',
    r'spell\.dict\..*',
    r'spell\.option\..*',
    r'trigger\.trigger\..*',
    r'weechat\.palette\..*',
    r'weechat\.proxy\..*',
    r'weechat\.bar\..*',
    r'weechat\.debug\..*',
    r'weechat\.notify\..*',
)

# completions to ignore
IGNORE_COMPLETIONS_ITEMS = (
    'docgen.*',
)


def translate(string):
    """Translate a string."""
    return _(string) if string else string


def escape(string):
    """Escape a string."""
    return string.replace('|', '\\|')


def sha256_file(filename, default=None):
    """Return SHA256 checksum of a file."""
    try:
        with open(filename, 'rb') as _file:
            checksum = hashlib.sha256(_file.read()).hexdigest()
    except IOError:
        checksum = default
    return checksum


class WeechatDoc():  # pylint: disable=too-few-public-methods
    """A class to read documentation from WeeChat API."""

    def __init__(self):
        pass

    def read_doc(self):
        """Get documentation from WeeChat API."""
        functions = sorted([
            func[0]
            for func in inspect.getmembers(self, predicate=inspect.isfunction)
            if func[0].startswith('_read_')
        ])
        return {
            function[6:]: getattr(self, function)()
            for function in functions
        }

    @staticmethod
    def _read_user_commands():
        """
        Get list of WeeChat/plugins commands as dictionary with 3 indexes:
        plugin, command, xxx.
        """
        commands = defaultdict(lambda: defaultdict(defaultdict))
        infolist = weechat.infolist_get('hook', '', 'command')
        while weechat.infolist_next(infolist):
            plugin = (weechat.infolist_string(infolist, 'plugin_name')
                      or 'weechat')
            if plugin in PLUGIN_LIST:
                command = weechat.infolist_string(infolist, 'command')
                if command == plugin or 'c' in PLUGIN_LIST[plugin]:
                    for key in ('description', 'args', 'args_description',
                                'completion'):
                        commands[plugin][command][key] = \
                            weechat.infolist_string(infolist, key)
        weechat.infolist_free(infolist)
        return commands

    @staticmethod
    def _read_user_options():
        """
        Get list of WeeChat/plugins config options as dictionary with
        4 indexes: config, section, option, xxx.
        """
        options = \
            defaultdict(lambda: defaultdict(lambda: defaultdict(defaultdict)))
        infolist = weechat.infolist_get('option', '', '')
        while weechat.infolist_next(infolist):
            full_name = weechat.infolist_string(infolist, 'full_name')
            if not re.search('|'.join(IGNORE_OPTIONS), full_name):
                config = weechat.infolist_string(infolist, 'config_name')
                if config in PLUGIN_LIST and 'o' in PLUGIN_LIST[config]:
                    section = weechat.infolist_string(infolist, 'section_name')
                    option = weechat.infolist_string(infolist, 'option_name')
                    for key in ('type', 'string_values', 'default_value',
                                'description'):
                        options[config][section][option][key] = \
                            weechat.infolist_string(infolist, key)
                    for key in ('min', 'max', 'null_value_allowed'):
                        options[config][section][option][key] = \
                            weechat.infolist_integer(infolist, key)
        weechat.infolist_free(infolist)
        return options

    @staticmethod
    def _read_api_infos():
        """
        Get list of WeeChat/plugins infos as dictionary with 3 indexes:
        plugin, name, xxx.
        """
        infos = defaultdict(lambda: defaultdict(defaultdict))
        infolist = weechat.infolist_get('hook', '', 'info')
        while weechat.infolist_next(infolist):
            info_name = weechat.infolist_string(infolist, 'info_name')
            plugin = (weechat.infolist_string(infolist, 'plugin_name')
                      or 'weechat')
            for key in ('description', 'args_description'):
                infos[plugin][info_name][key] = \
                    weechat.infolist_string(infolist, key)
        weechat.infolist_free(infolist)
        return infos

    @staticmethod
    def _read_api_infos_hashtable():
        """
        Get list of WeeChat/plugins infos (hashtable) as dictionary with
        3 indexes: plugin, name, xxx.
        """
        infos_hashtable = defaultdict(lambda: defaultdict(defaultdict))
        infolist = weechat.infolist_get('hook', '', 'info_hashtable')
        while weechat.infolist_next(infolist):
            info_name = weechat.infolist_string(infolist, 'info_name')
            plugin = (weechat.infolist_string(infolist, 'plugin_name')
                      or 'weechat')
            for key in ('description', 'args_description',
                        'output_description'):
                infos_hashtable[plugin][info_name][key] = \
                    weechat.infolist_string(infolist, key)
        weechat.infolist_free(infolist)
        return infos_hashtable

    @staticmethod
    def _read_api_infolists():
        """
        Get list of WeeChat/plugins infolists as dictionary with 3 indexes:
        plugin, name, xxx.
        """
        infolists = defaultdict(lambda: defaultdict(defaultdict))
        infolist = weechat.infolist_get('hook', '', 'infolist')
        while weechat.infolist_next(infolist):
            infolist_name = weechat.infolist_string(infolist, 'infolist_name')
            plugin = (weechat.infolist_string(infolist, 'plugin_name')
                      or 'weechat')
            for key in ('description', 'pointer_description',
                        'args_description'):
                infolists[plugin][infolist_name][key] = \
                    weechat.infolist_string(infolist, key)
        weechat.infolist_free(infolist)
        return infolists

    @staticmethod
    def _read_api_hdata():  # pylint: disable=too-many-locals
        """
        Get list of WeeChat/plugins hdata as dictionary with 3 indexes:
        plugin, name, xxx.
        """
        hdata = defaultdict(lambda: defaultdict(defaultdict))
        infolist = weechat.infolist_get('hook', '', 'hdata')
        while weechat.infolist_next(infolist):
            hdata_name = weechat.infolist_string(infolist, 'hdata_name')
            plugin = (weechat.infolist_string(infolist, 'plugin_name')
                      or 'weechat')
            hdata[plugin][hdata_name]['description'] = \
                weechat.infolist_string(infolist, 'description')
            variables = ''
            vars_update = ''
            lists = ''
            ptr_hdata = weechat.hdata_get(hdata_name)
            if ptr_hdata:
                hdata2 = []
                string = weechat.hdata_get_string(ptr_hdata, 'var_keys_values')
                if string:
                    for item in string.split(','):
                        key = item.split(':')[0]
                        var_offset = weechat.hdata_get_var_offset(
                            ptr_hdata,
                            key,
                        )
                        var_array_size = \
                            weechat.hdata_get_var_array_size_string(
                                ptr_hdata,
                                '',
                                key,
                            )
                        if var_array_size:
                            var_array_size = \
                                f', array_size: "{var_array_size}"'
                        var_hdata = weechat.hdata_get_var_hdata(ptr_hdata, key)
                        if var_hdata:
                            var_hdata = f', hdata: "{var_hdata}"'
                        type_string = weechat.hdata_get_var_type_string(
                            ptr_hdata,
                            key,
                        )
                        hdata2.append({
                            'offset': var_offset,
                            'text': f'_{key}_ ({type_string})',
                            'textlong': (f'_{key}_   ({type_string}'
                                         f'{var_array_size}{var_hdata})'),
                            'update': weechat.hdata_update(
                                ptr_hdata, '', {'__update_allowed': key}),
                        })
                    hdata2 = sorted(hdata2, key=itemgetter('offset'))
                    for item in hdata2:
                        variables += f'{item["textlong"]} +\n'
                        if item['update']:
                            vars_update += f'    {item["text"]} +\n'
                    if weechat.hdata_update(ptr_hdata, '',
                                            {'__create_allowed': ''}):
                        vars_update += '    _{hdata_update_create}_ +\n'
                    if weechat.hdata_update(ptr_hdata, '',
                                            {'__delete_allowed': ''}):
                        vars_update += '    _{hdata_update_delete}_ +\n'
                hdata[plugin][hdata_name]['vars'] = variables
                hdata[plugin][hdata_name]['vars_update'] = vars_update.rstrip()

                string = weechat.hdata_get_string(ptr_hdata, 'list_keys')
                if string:
                    list_lists = string.split(',')
                    lists_std = [lst for lst in list_lists
                                 if not lst.startswith('last_')]
                    lists_last = [lst for lst in list_lists
                                  if lst.startswith('last_')]
                    for item in sorted(lists_std) + sorted(lists_last):
                        lists += f'_{item}_ +\n'
                hdata[plugin][hdata_name]['lists'] = lists
        weechat.infolist_free(infolist)
        return hdata

    @staticmethod
    def _read_api_completions():
        """
        Get list of WeeChat/plugins completions as dictionary with 3 indexes:
        plugin, item, xxx.
        """
        completions = defaultdict(lambda: defaultdict(defaultdict))
        infolist = weechat.infolist_get('hook', '', 'completion')
        while weechat.infolist_next(infolist):
            completion_item = weechat.infolist_string(infolist,
                                                      'completion_item')
            if not re.search('|'.join(IGNORE_COMPLETIONS_ITEMS),
                             completion_item):
                plugin = (weechat.infolist_string(infolist, 'plugin_name')
                          or 'weechat')
                completions[plugin][completion_item]['description'] = \
                    weechat.infolist_string(infolist, 'description')
        weechat.infolist_free(infolist)
        return completions

    @staticmethod
    def _read_api_url_options():
        """
        Get list of URL options as list of dictionaries.
        """
        url_options = []
        infolist = weechat.infolist_get('url_options', '', '')
        while weechat.infolist_next(infolist):
            url_options.append({
                'name': weechat.infolist_string(infolist, 'name').lower(),
                'option': weechat.infolist_integer(infolist, 'option'),
                'type': weechat.infolist_string(infolist, 'type'),
                'constants': weechat.infolist_string(
                    infolist, 'constants').lower().replace(',', ', ')
            })
        weechat.infolist_free(infolist)
        return url_options

    @staticmethod
    def _read_user_default_aliases():
        """
        Get list of default aliases as list of dictionaries.
        """
        default_aliases = []
        infolist = weechat.infolist_get('alias_default', '', '')
        while weechat.infolist_next(infolist):
            default_aliases.append({
                'name': '/' + weechat.infolist_string(infolist, 'name'),
                'command': '/' + weechat.infolist_string(infolist, 'command'),
                'completion': weechat.infolist_string(infolist, 'completion'),
            })
        weechat.infolist_free(infolist)
        return default_aliases

    @staticmethod
    def _read_user_irc_colors():
        """
        Get list of IRC colors as list of dictionaries.
        """
        irc_colors = []
        infolist = weechat.infolist_get('irc_color_weechat', '', '')
        while weechat.infolist_next(infolist):
            irc_colors.append({
                'color_irc': weechat.infolist_string(infolist, 'color_irc'),
                'color_weechat': weechat.infolist_string(infolist,
                                                         'color_weechat'),
            })
        weechat.infolist_free(infolist)
        return irc_colors

    @staticmethod
    def _read_api_plugins_priority():
        """
        Get priority of default WeeChat plugins as a dictionary.
        """
        plugins_priority = {}
        infolist = weechat.infolist_get('plugin', '', '')
        while weechat.infolist_next(infolist):
            name = weechat.infolist_string(infolist, 'name')
            priority = weechat.infolist_integer(infolist, 'priority')
            if priority in plugins_priority:
                plugins_priority[priority].append(name)
            else:
                plugins_priority[priority] = [name]
        weechat.infolist_free(infolist)
        return plugins_priority


class AutogenDoc():
    """A class to write auto-generated doc files."""

    def __init__(self, weechat_doc, doc_directory, locale):
        """Initialize auto-generated doc file."""
        self.doc_directory = doc_directory
        self.locale = locale
        self.count_files = 0
        self.count_updated = 0
        self.filename = None
        self.filename_tmp = None
        self._file = None
        self.install_translations()
        self.write_autogen_files(weechat_doc)

    def install_translations(self):
        """Install translations."""
        trans = gettext.translation(
            'weechat',
            weechat.info_get('weechat_localedir', ''),
            languages=[f'{self.locale}.UTF-8'],
            fallback=True,
        )
        trans.install()

    def open_file(self, name):
        """Open temporary auto-generated file."""
        self.filename = os.path.join(
            self.doc_directory,
            self.locale[:2],
            'includes',
            f'autogen_{name}.{self.locale[:2]}.adoc',
        )
        self.filename_tmp = f'{self.filename}.tmp'
        self._file = open(self.filename_tmp, 'w')

    def write_autogen_files(self, weechat_doc):
        """Write auto-generated files."""
        for name, doc in weechat_doc.items():
            self.open_file(name)
            self.write_autogen_file(name, doc)
            self.update_autogen_file()

    def write_autogen_file(self, name, doc):
        """Write auto-generated file."""
        self.write('//')
        self.write('// This file is auto-generated by script docgen.py.')
        self.write('// DO NOT EDIT BY HAND!')
        self.write('//')
        getattr(self, f'_write_{name}')(doc)

    def write(self, *args):
        """Write a line in auto-generated doc file."""
        if args:
            if len(args) > 1:
                self._file.write(args[0] % args[1:])
            else:
                self._file.write(args[0])
        self._file.write('\n')

    def update_autogen_file(self):
        """Update doc file if needed (if content has changed)."""
        self.count_files += 1
        # close temp file
        self._file.close()
        sha_old = sha256_file(self.filename, 'old')
        sha_new = sha256_file(self.filename_tmp, 'new')
        # compare checksums
        if sha_old != sha_new:
            # update doc file
            if os.path.exists(self.filename):
                os.unlink(self.filename)
            os.rename(self.filename_tmp, self.filename)
            self.count_updated += 1
        else:
            os.unlink(self.filename_tmp)

    def __str__(self):
        """Get status string."""
        if self.count_updated > 0:
            color_count = weechat.color('yellow')
            color_updated = weechat.color('green')
            color_reset = weechat.color('reset')
            str_updated = (f', {color_count}{self.count_updated} '
                           f'{color_updated}updated{color_reset}')
        else:
            str_updated = ''
        return f'{self.locale}: {self.count_files} files{str_updated}'

    def _write_user_commands(self, commands):
        """Write commands."""
        for plugin in commands:
            self.write()
            self.write(f'// tag::{plugin}_commands[]')
            for i, command in enumerate(sorted(commands[plugin])):
                if i > 0:
                    self.write()
                _cmd = commands[plugin][command]
                args = translate(_cmd['args'])
                args_formats = args.split(' || ')
                desc = translate(_cmd['description'])
                args_desc = translate(_cmd['args_description'])
                self.write(f'[[command_{plugin}_{command}]]')
                self.write(f'* `+{command}+`: {desc}\n')
                self.write('----')
                prefix = '/' + command + '  '
                if args_formats != ['']:
                    for fmt in args_formats:
                        self.write(prefix + fmt)
                        prefix = ' ' * len(prefix)
                if args_desc:
                    self.write()
                    self.write(args_desc)
                self.write('----')
            self.write(f'// end::{plugin}_commands[]')

    # pylint: disable=too-many-locals,too-many-branches
    def _write_user_options(self, options):
        """Write config options."""
        for config in options:
            self.write()
            self.write(f'// tag::{config}_options[]')
            i = 0
            for section in sorted(options[config]):
                for option in sorted(options[config][section]):
                    if i > 0:
                        self.write()
                    i += 1
                    _opt = options[config][section][option]
                    opt_type = _opt['type']
                    string_values = _opt['string_values']
                    default_value = _opt['default_value']
                    opt_min = _opt['min']
                    opt_max = _opt['max']
                    null_value_allowed = _opt['null_value_allowed']
                    desc = translate(_opt['description'])
                    type_nls = translate(opt_type)
                    values = ''
                    if opt_type == 'boolean':
                        values = 'on, off'
                    elif opt_type == 'integer':
                        if string_values:
                            values = string_values.replace('|', ', ')
                        else:
                            values = f'{opt_min} .. {opt_max}'
                    elif opt_type == 'string':
                        if opt_max <= 0:
                            values = _('any string')
                        elif opt_max == 1:
                            values = _('any char')
                        elif opt_max > 1:
                            values = (_('any string')
                                      + '(' + _('max chars') + ': '
                                      + opt_max + ')')
                        else:
                            values = _('any string')
                        default_value = ('"%s"' %
                                         default_value.replace('"', '\\"'))
                    elif opt_type == 'color':
                        values = _('a WeeChat color name (default, black, '
                                   '(dark)gray, white, (light)red, '
                                   '(light)green, brown, yellow, (light)blue, '
                                   '(light)magenta, (light)cyan), a terminal '
                                   'color number or an alias; attributes are '
                                   'allowed before color (for text color '
                                   'only, not background): \"*\" for bold, '
                                   '\"!\" for reverse, \"/\" for italic, '
                                   '\"_\" for underline')
                    self.write(f'* [[option_{config}.{section}.{option}]] '
                               f'*{config}.{section}.{option}*')
                    self.write('** %s: pass:none[%s]',
                               _('description'), desc.replace(']', '\\]'))
                    self.write('** %s: %s', _('type'), type_nls)
                    self.write('** %s: %s', _('values'), values)
                    self.write('** %s: `+%s+`',
                               _('default value'), default_value)
                    if null_value_allowed:
                        self.write('** %s',
                                   _('undefined value allowed (null)'))
            self.write(f'// end::{config}_options[]')

    def _write_user_default_aliases(self, default_aliases):
        """Write default aliases."""
        self.write()
        self.write('// tag::default_aliases[]')
        self.write('[width="100%",cols="2m,5m,5",options="header"]')
        self.write('|===')
        self.write('| %s | %s | %s\n',
                   _('Alias'), _('Command'), _('Completion'))
        for alias in default_aliases:
            self.write('| %s | %s | %s',
                       escape(alias['name']),
                       escape(alias['command']),
                       escape(alias['completion'] or '-'))
        self.write('|===')
        self.write('// end::default_aliases[]')

    def _write_user_irc_colors(self, irc_colors):
        """Write IRC colors."""
        self.write()
        self.write('// tag::irc_colors[]')
        self.write('[width="50%",cols="^2m,3",options="header"]')
        self.write('|===')
        self.write('| %s | %s\n', _('IRC color'), _('WeeChat color'))
        for color in irc_colors:
            self.write('| %s | %s',
                       escape(color['color_irc']),
                       escape(color['color_weechat']))
        self.write('|===')
        self.write('// end::irc_colors[]')

    def _write_api_infos(self, infos):
        """Write infos."""
        self.write()
        self.write('// tag::infos[]')
        self.write('[width="100%",cols="^1,^2,6,6",options="header"]')
        self.write('|===')
        self.write('| %s | %s | %s | %s\n',
                   _('Plugin'), _('Name'), _('Description'), _('Arguments'))
        for plugin in sorted(infos):
            for info in sorted(infos[plugin]):
                _inf = infos[plugin][info]
                desc = translate(_inf['description'])
                args_desc = translate(_inf['args_description'] or '-')
                self.write('| %s | %s | %s | %s\n',
                           escape(plugin), escape(info), escape(desc),
                           escape(args_desc))
        self.write('|===')
        self.write('// end::infos[]')

    def _write_api_infos_hashtable(self, infos_hashtable):
        """Write infos hashtable."""
        self.write()
        self.write('// tag::infos_hashtable[]')
        self.write('[width="100%",cols="^1,^2,6,6,8",options="header"]')
        self.write('|===')
        self.write('| %s | %s | %s | %s | %s\n',
                   _('Plugin'), _('Name'), _('Description'),
                   _('Hashtable (input)'), _('Hashtable (output)'))
        for plugin in sorted(infos_hashtable):
            for info in sorted(infos_hashtable[plugin]):
                _inh = infos_hashtable[plugin][info]
                desc = translate(_inh['description'])
                args_desc = translate(_inh['args_description'])
                output_desc = translate(_inh['output_description']) or '-'
                self.write('| %s | %s | %s | %s | %s\n',
                           escape(plugin), escape(info), escape(desc),
                           escape(args_desc), escape(output_desc))
        self.write('|===')
        self.write('// end::infos_hashtable[]')

    def _write_api_infolists(self, infolists):
        """Write infolists."""
        self.write()
        self.write('// tag::infolists[]')
        self.write('[width="100%",cols="^1,^2,5,5,5",options="header"]')
        self.write('|===')
        self.write('| %s | %s | %s | %s | %s\n',
                   _('Plugin'), _('Name'), _('Description'), _('Pointer'),
                   _('Arguments'))
        for plugin in sorted(infolists):
            for infolist in sorted(infolists[plugin]):
                _inl = infolists[plugin][infolist]
                desc = translate(_inl['description'])
                pointer_desc = translate(_inl['pointer_description']) or '-'
                args_desc = translate(_inl['args_description']) or '-'
                self.write('| %s | %s | %s | %s | %s\n',
                           escape(plugin), escape(infolist), escape(desc),
                           escape(pointer_desc), escape(args_desc))
        self.write('|===')
        self.write('// end::infolists[]')

    def _write_api_hdata(self, hdata):
        """Write hdata."""
        self.write()
        self.write('// tag::hdata[]')
        self.write(':hdata_update_create: __create')
        self.write(':hdata_update_delete: __delete')
        self.write('[width="100%",cols="^1,^2,2,2,5",options="header"]')
        self.write('|===')
        self.write('| %s | %s | %s | %s | %s\n',
                   _('Plugin'), _('Name'), _('Description'), _('Lists'),
                   _('Variables'))
        for plugin in sorted(hdata):
            for hdata_name in sorted(hdata[plugin]):
                _hda = hdata[plugin][hdata_name]
                anchor = f'hdata_{hdata_name}'
                desc = translate(_hda['description'])
                variables = _hda['vars']
                vars_update = _hda['vars_update']
                lists = _hda['lists']
                self.write(f'| {escape(plugin)}')
                self.write(f'| [[{escape(anchor)}]]<<{escape(anchor)},'
                           f'{escape(hdata_name)}>>')
                self.write(f'| {escape(desc)}')
                str_lists = escape(lists) if lists else '-'
                self.write(f'| {str_lists}')
                self.write(f'| {escape(variables)}')
                if vars_update:
                    self.write('*%s* +\n%s',
                               _('Update allowed:'), escape(vars_update))
                self.write()
        self.write('|===')
        self.write('// end::hdata[]')

    def _write_api_completions(self, completions):
        """Write completions."""
        self.write()
        self.write('// tag::completions[]')
        self.write('[width="100%",cols="^1,^2,7",options="header"]')
        self.write('|===')
        self.write('| %s | %s | %s\n',
                   _('Plugin'), _('Name'), _('Description'))
        for plugin in sorted(completions):
            for completion_item in sorted(completions[plugin]):
                _cmp = completions[plugin][completion_item]
                desc = translate(_cmp['description'])
                self.write('| %s | %s | %s\n',
                           escape(plugin), escape(completion_item),
                           escape(desc))
        self.write('|===')
        self.write('// end::completions[]')

    def _write_api_url_options(self, url_options):
        """Write URL options."""
        self.write()
        self.write('// tag::url_options[]')
        self.write('[width="100%",cols="2,^1,7",options="header"]')
        self.write('|===')
        self.write('| %s | %s ^(1)^ | %s ^(2)^\n',
                   _('Option'), _('Type'), _('Constants'))
        for option in url_options:
            constants = option['constants']
            if constants:
                constants = ' ' + constants
            self.write('| %s | %s |%s\n',
                       escape(option['name']), escape(option['type']),
                       escape(constants))
        self.write('|===')
        self.write('// end::url_options[]')

    def _write_api_plugins_priority(self, plugins_priority):
        """Write plugins priority."""
        self.write()
        self.write('// tag::plugins_priority[]')
        for priority in sorted(plugins_priority, reverse=True):
            plugins = ', '.join(sorted(plugins_priority[priority]))
            self.write('. %s (%s)', escape(plugins), priority)
        self.write('// end::plugins_priority[]')


def docgen_cmd_cb(data, buf, args):
    """Callback for /docgen command."""
    doc_directory = data
    locales = args.split(' ') if args else sorted(LOCALE_LIST)

    weechat_doc = WeechatDoc().read_doc()

    weechat.prnt('', '-' * 75)

    for locale in locales:
        autogen = AutogenDoc(weechat_doc, doc_directory, locale)
        weechat.prnt('', f'docgen: {autogen}')

    weechat.prnt('', '-' * 75)

    return weechat.WEECHAT_RC_OK


def docgen_completion_cb(data, completion_item, buf, completion):
    """Callback for completion."""
    for locale in LOCALE_LIST:
        weechat.completion_list_add(completion, locale, 0,
                                    weechat.WEECHAT_LIST_POS_SORT)
    return weechat.WEECHAT_RC_OK


if __name__ == '__main__' and IMPORT_OK:
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION,
                        SCRIPT_LICENSE, SCRIPT_DESC, '', ''):
        weechat.hook_command(
            SCRIPT_COMMAND,
            'Documentation generator.',
            '[locales]',
            'locales: list of locales to build (by default build all locales)',
            '%(docgen_locales)|%*',
            'docgen_cmd_cb',
            os.path.dirname(__file__),
        )
        weechat.hook_completion(
            'docgen_locales',
            'locales for docgen',
            'docgen_completion_cb',
            '',
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020 Sébastien Helleu <flashcode@flashtux.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""
Unparse AST tree to generate scripts in all supported languages
(Python, Perl, Ruby, ...).
"""

# pylint: disable=too-many-lines,unnecessary-pass

from __future__ import print_function
import argparse
import ast
import inspect
import os
import select
try:
    from StringIO import StringIO  # python 2
except ImportError:
    from io import StringIO  # python 3
import sys

sys.dont_write_bytecode = True


class UnparsePython(object):
    """
    Unparse AST to generate Python script code.

    This class is inspired from unparse.py in cpython:
    https://github.com/python/cpython/blob/master/Tools/parser/unparse.py

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def __init__(self, output=sys.stdout):
        self.output = output
        self.indent_string = ' ' * 4
        self._indent_level = 0
        self._prefix = []  # not used in Python, only in other languages
        self.binop = {
            'Add': '+',
            'Sub': '-',
            'Mult': '*',
            'MatMult': '@',
            'Div': '/',
            'Mod': '%',
            'LShift': '<<',
            'RShift': '>>',
            'BitOr': '|',
            'BitXor': '^',
            'BitAnd': '&',
            'FloorDiv': '//',
            'Pow': '**',
        }
        self.unaryop = {
            'Invert': '~',
            'Not': 'not',
            'UAdd': '+',
            'USub': '-',
        }
        self.cmpop = {
            'Eq': '==',
            'NotEq': '!=',
            'Lt': '<',
            'LtE': '<=',
            'Gt': '>',
            'GtE': '>=',
            'Is': 'is',
            'IsNot': 'is not',
            'In': 'in',
            'NotIn': 'not in',
        }

    def fill(self, string=''):
        """Add a new line and an indented string."""
        self.add('\n%s%s' % (self.indent_string * self._indent_level, string))

    def indent(self):
        """Indent code."""
        self._indent_level += 1

    def unindent(self):
        """Unindent code."""
        self._indent_level -= 1

    def prefix(self, prefix):
        """Add or remove a prefix from list."""
        if prefix:
            self._prefix.append(prefix)
        else:
            self._prefix.pop()

    def add(self, *args):
        """Add string/node(s) to the output file."""
        for arg in args:
            if callable(arg):
                arg()
            elif isinstance(arg, tuple):
                arg[0](*arg[1:])
            elif isinstance(arg, list):
                for item in arg:
                    self.add(item)
            elif isinstance(arg, ast.AST):
                method = getattr(
                    self, '_ast_%s' % arg.__class__.__name__.lower(),
                    None)
                if method is None:
                    raise NotImplementedError(arg)
                method(arg)
            elif isinstance(arg, str):
                self.output.write(arg)

    @staticmethod
    def make_list(values, sep=', '):
        """Add multiple values using a custom method and separator."""
        result = []
        for value in values:
            if result:
                result.append(sep)
            result.append(value)
        return result

    def is_bool(self, node):  # pylint: disable=no-self-use
        """Check if the node is a boolean."""
        return isinstance(node, ast.Name) and node.id in ('False', 'True')

    def is_number(self, node):  # pylint: disable=no-self-use
        """Check if the node is a number."""
        # in python 2, number -1 is Num(n=-1)
        # in Python 3, number -1 is UnaryOp(op=USub(), operand=Num(n=1))
        return (isinstance(node, ast.Num) or
                (isinstance(node, ast.UnaryOp) and
                 isinstance(node.op, (ast.UAdd, ast.USub))))

    def _ast_alias(self, node):
        """Add an AST alias in output."""
        # ignore alias
        pass

    def _ast_arg(self, node):
        """Add an AST arg in output."""
        self.add('%s%s' % (self._prefix[-1] if self._prefix else '',
                           node.arg))

    def _ast_assign(self, node):
        """Add an AST Assign in output."""
        self.add(
            self.fill,
            [[target, ' = '] for target in node.targets],
            node.value,
        )

    def _ast_attribute(self, node):
        """Add an AST Attribute in output."""
        self.add(node.value, '.', node.attr)

    def _ast_binop(self, node):
        """Add an AST BinOp in output."""
        self.add(
            node.left,
            ' %s ' % self.binop[node.op.__class__.__name__],
            node.right,
        )

    def _ast_call(self, node):
        """Add an AST Call in output."""
        self.add(
            node.func,
            '(',
            self.make_list(node.args),
            ')',
        )

    def _ast_compare(self, node):
        """Add an AST Compare in output."""
        self.add(node.left)
        for operator, comparator in zip(node.ops, node.comparators):
            self.add(
                ' %s ' % self.cmpop[operator.__class__.__name__],
                comparator,
            )

    def _ast_constant(self, node):
        """Add an AST Constant in output."""
        self.add(repr(node.s))

    def _ast_dict(self, node):
        """Add an AST Dict in output."""
        self.add(
            '{',
            self.make_list([[key, ': ', value]
                            for key, value in zip(node.keys, node.values)]),
            '}',
        )

    def _ast_expr(self, node):
        """Add an AST Expr in output."""
        if not isinstance(node.value, ast.Str):  # ignore docstrings
            self.add(
                self.fill,
                node.value,
            )

    def _ast_functiondef(self, node):
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            self.fill if self._indent_level == 0 else None,
            'def %s(' % node.name,
            self.make_list(node.args.args),
            '):',
            self.indent,
            node.body,
            self.unindent,
        )

    def _ast_if(self, node):
        """Add an AST If in output."""
        self.add(
            self.fill,
            'if ',
            node.test,
            ':',
            self.indent,
            node.body,
            self.unindent,
        )
        if node.orelse:
            self.add(
                self.fill,
                'else:',
                self.indent,
                node.orelse,
                self.unindent,
            )

    def _ast_import(self, node):
        """Add an AST Import in output."""
        # ignore import
        pass

    def _ast_module(self, node):
        """Add an AST Module in output."""
        self.add(node.body)

    def _ast_name(self, node):
        """Add an AST Name in output."""
        self.add('%s%s' % (self._prefix[-1] if self._prefix else '',
                           node.id))

    def _ast_num(self, node):
        """Add an AST Num in output."""
        # note: deprecated since Python 3.8, replaced by ast.Constant
        self.add(repr(node.n))

    def _ast_pass(self, node):  # pylint: disable=unused-argument
        """Add an AST Pass in output."""
        self.fill('pass')

    def _ast_return(self, node):
        """Add an AST Return in output."""
        self.fill('return')
        if node.value:
            self.add(' ', node.value)

    def _ast_str(self, node):
        """Add an AST Str in output."""
        # note: deprecated since Python 3.8, replaced by ast.Constant
        self.add(repr(node.s))

    def _ast_tuple(self, node):
        """Add an AST Tuple in output."""
        self.add(
            '(',
            self.make_list(node.elts),
            ',' if len(node.elts) == 1 else None,
            ')',
        )

    def _ast_unaryop(self, node):
        """Add an AST UnaryOp in output."""
        self.add(
            '(',
            self.unaryop[node.op.__class__.__name__],
            ' ',
            node.operand,
            ')',
        )


class UnparsePerl(UnparsePython):
    """
    Unparse AST to generate Perl script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def _ast_assign(self, node):
        """Add an AST Assign in output."""
        self.add(
            self.fill,
            (self.prefix, '%' if isinstance(node.value, ast.Dict) else '$'),
            [[target, ' = '] for target in node.targets],
            (self.prefix, None),
            node.value,
            ';',
        )

    def _ast_attribute(self, node):
        """Add an AST Attribute in output."""
        saved_prefix = self._prefix
        self._prefix = []
        self.add(node.value, '::', node.attr)
        self._prefix = saved_prefix

    def _ast_binop(self, node):
        """Add an AST BinOp in output."""
        if isinstance(node.op, ast.Add) and \
                (not self.is_number(node.left) or
                 not self.is_number(node.right)):
            str_op = '.'
        else:
            str_op = self.binop[node.op.__class__.__name__]
        self.add(
            (self.prefix, '$'),
            node.left,
            ' %s ' % str_op,
            node.right,
            (self.prefix, None),
        )

    def _ast_call(self, node):
        """Add an AST Call in output."""
        self.add(
            node.func,
            '(',
            (self.prefix, '$'),
            self.make_list(node.args),
            (self.prefix, None),
            ')',
        )

    def _ast_compare(self, node):
        """Add an AST Compare in output."""
        self.add(node.left)
        for operator, comparator in zip(node.ops, node.comparators):
            if isinstance(operator, (ast.Eq, ast.NotEq)) and \
                    not self.is_number(node.left) and \
                    not self.is_bool(node.left) and \
                    not self.is_number(comparator) and \
                    not self.is_bool(comparator):
                custom_cmpop = {
                    'Eq': 'eq',
                    'NotEq': 'ne',
                }
            else:
                custom_cmpop = self.cmpop
            self.add(
                ' %s ' % custom_cmpop[operator.__class__.__name__],
                comparator,
            )

    def _ast_constant(self, node):
        """Add an AST Constant in output."""
        if isinstance(node.value, str):
            self.add('"%s"' % node.s.replace('$', '\\$').replace('@', '\\@'))
        else:
            self.add(repr(node.s))

    def _ast_dict(self, node):
        """Add an AST Dict in output."""
        self.add(
            '{',
            self.make_list([[key, ' => ', value]
                            for key, value in zip(node.keys, node.values)]),
            '}',
        )

    def _ast_expr(self, node):
        """Add an AST Expr in output."""
        if not isinstance(node.value, ast.Str):  # ignore docstrings
            self.add(
                self.fill,
                node.value,
                ';',
            )

    def _ast_functiondef(self, node):
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            'sub %s' % node.name,
            self.fill,
            '{',
            self.indent,
        )
        if node.args.args:
            self.add(
                self.fill,
                'my (',
                (self.prefix, '$'),
                self.make_list(node.args.args),
                (self.prefix, None),
                ') = @_;',
            )
        self.add(
            node.body,
            self.unindent,
            self.fill,
            '}',
        )

    def _ast_if(self, node):
        """Add an AST If in output."""
        self.add(
            self.fill,
            'if (',
            (self.prefix, '$'),
            node.test,
            (self.prefix, None),
            ')',
            self.fill,
            '{',
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            '}',
        )
        if node.orelse:
            self.add(
                self.fill,
                'else',
                self.fill,
                '{',
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                '}',
            )

    def _ast_pass(self, node):
        """Add an AST Pass in output."""
        pass

    def _ast_return(self, node):
        """Add an AST Return in output."""
        self.fill('return')
        if node.value:
            self.add(
                ' ',
                (self.prefix,
                 '%' if isinstance(node.value, ast.Dict) else '$'),
                node.value,
                (self.prefix, None),
                ';',
            )

    def _ast_str(self, node):
        """Add an AST Str in output."""
        # note: deprecated since Python 3.8, replaced by ast.Constant
        self.add('"%s"' % node.s.replace('$', '\\$').replace('@', '\\@'))


class UnparseRuby(UnparsePython):
    """
    Unparse AST to generate Ruby script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def _ast_attribute(self, node):
        """Add an AST Attribute in output."""
        self.add(
            node.value,
            '::' if node.attr.startswith('WEECHAT_') else '.',
            node.attr,
        )

    def _ast_dict(self, node):
        """Add an AST Dict in output."""
        self.add(
            'Hash[',
            self.make_list([[key, ' => ', value]
                            for key, value in zip(node.keys, node.values)]),
            ']',
        )

    def _ast_functiondef(self, node):
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            'def %s' % node.name,
        )
        if node.args.args:
            self.add(
                '(',
                self.make_list(node.args.args),
                ')',
            )
        self.add(
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            'end',
        )

    def _ast_if(self, node):
        """Add an AST If in output."""
        self.add(
            self.fill,
            'if ',
            node.test,
            self.indent,
            node.body,
            self.unindent,
        )
        if node.orelse:
            self.add(
                self.fill,
                'else',
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                'end',
            )

    def _ast_pass(self, node):
        """Add an AST Pass in output."""
        pass

    def _ast_str(self, node):
        """Add an AST Str in output."""
        # note: deprecated since Python 3.8, replaced by ast.Constant
        self.add('"%s"' % node.s)


class UnparseLua(UnparsePython):
    """
    Unparse AST to generate Lua script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def __init__(self, *args, **kwargs):
        super(UnparseLua, self).__init__(*args, **kwargs)
        self.cmpop = {
            'Eq': '==',
            'NotEq': '~=',
            'Lt': '<',
            'LtE': '<=',
            'Gt': '>',
            'GtE': '>=',
        }

    def _ast_binop(self, node):
        """Add an AST BinOp in output."""
        if isinstance(node.op, ast.Add) and \
                (not self.is_number(node.left) or
                 not self.is_number(node.right)):
            str_op = '..'
        else:
            str_op = self.binop[node.op.__class__.__name__]
        self.add(
            node.left,
            ' %s ' % str_op,
            node.right,
        )

    def _ast_dict(self, node):
        """Add an AST Dict in output."""
        self.add(
            '{',
            self.make_list([
                ['[', key, ']', '=', value]
                for key, value in zip(node.keys, node.values)]),
            '}',
        )

    def _ast_functiondef(self, node):
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            'function %s' % node.name,
        )
        self.add(
            '(',
            self.make_list(node.args.args),
            ')',
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            'end',
        )

    def _ast_if(self, node):
        """Add an AST If in output."""
        self.add(
            self.fill,
            'if ',
            node.test,
            ' then',
            self.indent,
            node.body,
            self.unindent,
        )
        if node.orelse:
            self.add(
                self.fill,
                'else',
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                'end',
            )

    def _ast_pass(self, node):
        """Add an AST Pass in output."""
        pass


class UnparseTcl(UnparsePython):
    """
    Unparse AST to generate Tcl script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def __init__(self, *args, **kwargs):
        super(UnparseTcl, self).__init__(*args, **kwargs)
        self._call = 0

    def _ast_assign(self, node):
        """Add an AST Assign in output."""
        self.add(
            self.fill,
            'set ',
            node.targets[0],
            ' ',
            '[' if not isinstance(node.value, ast.Str) else '',
            node.value,
            ']' if not isinstance(node.value, ast.Str) else '',
        )

    def _ast_attribute(self, node):
        """Add an AST Attribute in output."""
        saved_prefix = self._prefix
        self._prefix = []
        if node.attr.startswith('WEECHAT_'):
            self.add('$::')
        self.add(node.value, '::', node.attr)
        self._prefix = saved_prefix

    def _ast_binop(self, node):
        """Add an AST BinOp in output."""
        self.add(
            '[join [list ',
            (self.prefix, '$'),
            node.left,
            ' ',
            node.right,
            (self.prefix, None),
            '] ""]',
        )

    def _ast_call(self, node):
        """Add an AST Call in output."""
        if self._call:
            self.add('[')
        self._call += 1
        self.add(
            node.func,
            ' ' if node.args else None,
            (self.prefix, '$'),
            self.make_list(node.args, sep=' '),
            (self.prefix, None),
        )
        self._call -= 1
        if self._call:
            self.add(']')

    def _ast_compare(self, node):
        """Add an AST Compare in output."""
        self.prefix('$')
        if self._call:
            self.add('[expr {')
        self.add(node.left)
        for operator, comparator in zip(node.ops, node.comparators):
            if isinstance(operator, (ast.Eq, ast.NotEq)) and \
                    not self.is_number(node.left) and \
                    not self.is_bool(node.left) and \
                    not self.is_number(comparator) and \
                    not self.is_bool(comparator):
                custom_cmpop = {
                    'Eq': 'eq',
                    'NotEq': 'ne',
                }
            else:
                custom_cmpop = self.cmpop
            self.add(
                ' %s ' % custom_cmpop[operator.__class__.__name__],
                comparator,
            )
        if self._call:
            self.add('}]')
        self.prefix(None)

    def _ast_constant(self, node):
        """Add an AST Constant in output."""
        if isinstance(node.value, str):
            self.add('"%s"' % node.s.replace('$', '\\$'))
        else:
            self.add(repr(node.s))

    def _ast_dict(self, node):
        """Add an AST Dict in output."""
        self.add(
            '[dict create ',
            self.make_list([[key, ' ', value]
                            for key, value in zip(node.keys, node.values)],
                           sep=' '),
            ']',
        )

    def _ast_functiondef(self, node):
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            'proc %s {' % node.name,
            (self.make_list(node.args.args, sep=' ')
             if node.args.args else None),
            '} {',
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            '}',
        )

    def _ast_if(self, node):
        """Add an AST If in output."""
        self.add(
            self.fill,
            'if {',
            (self.prefix, '$'),
            node.test,
            (self.prefix, None),
            '} {',
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            '}',
        )
        if node.orelse:
            self.add(
                ' else {',
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                '}',
            )

    def _ast_pass(self, node):
        """Add an AST Pass in output."""
        pass

    def _ast_return(self, node):
        """Add an AST Return in output."""
        self.fill('return')
        if node.value:
            self.add(
                ' ',
                (self.prefix, '$'),
                node.value,
                (self.prefix, None),
            )

    def _ast_str(self, node):
        """Add an AST Str in output."""
        # note: deprecated since Python 3.8, replaced by ast.Constant
        self.add('"%s"' % node.s.replace('$', '\\$'))


class UnparseGuile(UnparsePython):
    """
    Unparse AST to generate Guile script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def __init__(self, *args, **kwargs):
        super(UnparseGuile, self).__init__(*args, **kwargs)
        self.cmpop = {
            'Eq': '=',
            'NotEq': '<>',
            'Lt': '<',
            'LtE': '<=',
            'Gt': '>',
            'GtE': '>=',
        }
        self._call = 0
        self._let = 0

    def _ast_assign(self, node):
        """Add an AST Assign in output."""
        self.add(
            self.fill,
            '(let ((',
            node.targets[0],
            ' ',
            node.value,
            '))',
            self.indent,
            self.fill,
            '(begin',
            self.indent,
        )
        self._let += 1

    def _ast_attribute(self, node):
        """Add an AST Attribute in output."""
        self.add(node.value, ':', node.attr)

    def _ast_binop(self, node):
        """Add an AST BinOp in output."""
        if isinstance(node.op, ast.Add) and \
                (isinstance(node.left, (ast.Name, ast.Str)) or
                 isinstance(node.right, (ast.Name, ast.Str))):
            self.add(
                '(string-append ',
                node.left,
                ' ',
                node.right,
                ')',
            )
        else:
            self.add(
                node.left,
                ' %s ' % self.binop[node.op.__class__.__name__],
                node.right,
            )

    def _ast_call(self, node):
        """Add an AST Call in output."""
        self._call += 1
        self.add(
            '(',
            node.func,
            ' ' if node.args else None,
            self.make_list(node.args, sep=' '),
            ')',
        )
        self._call -= 1

    def _ast_compare(self, node):
        """Add an AST Compare in output."""
        for operator, comparator in zip(node.ops, node.comparators):
            if isinstance(operator, (ast.Eq, ast.NotEq)) and \
                    not self.is_number(node.left) and \
                    not self.is_bool(node.left) and \
                    not self.is_number(comparator) and \
                    not self.is_bool(comparator):
                prefix = 'string'
            else:
                prefix = ''
            self.add(
                '(%s%s ' % (prefix, self.cmpop[operator.__class__.__name__]),
                node.left,
                ' ',
                comparator,
                ')',
            )

    def _ast_constant(self, node):
        """Add an AST Constant in output."""
        if isinstance(node.s, str):
            self.add('"%s"' % node.s)
        else:
            self.add(repr(node.s))

    def _ast_dict(self, node):
        """Add an AST Dict in output."""
        self.add(
            '\'(',
            self.make_list([['(', key, ' ', value, ')']
                            for key, value in zip(node.keys, node.values)],
                           sep=' '),
            ')',
        )

    def _ast_functiondef(self, node):
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            '(define (%s' % node.name,
            ' ' if node.args.args else None,
            (self.make_list(node.args.args, sep=' ')
             if node.args.args else None),
            ')',
            self.indent,
            node.body,
        )
        while self._let > 0:
            self.add(
                self.unindent,
                self.fill,
                ')',
                self.unindent,
                self.fill,
                ')',
            )
            self._let -= 1
        self.add(
            self.unindent,
            self.fill,
            ')',
        )

    def _ast_if(self, node):
        """Add an AST If in output."""
        self.add(
            self.fill,
            '(if '
            '' if isinstance(node.test, ast.Name) else '(',
            node.test,
            '' if isinstance(node.test, ast.Name) else ')',
            self.indent,
            self.fill,
            '(begin',
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            ')',
            self.unindent,
        )
        if node.orelse:
            self.add(
                self.indent,
                self.fill,
                '(begin',
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                ')',
                self.unindent,
            )
        self.add(self.fill, ')')

    def _ast_pass(self, node):
        """Add an AST Pass in output."""
        pass

    def _ast_return(self, node):
        """Add an AST Return in output."""
        if node.value:
            self.add(self.fill, node.value)

    def _ast_str(self, node):
        """Add an AST Str in output."""
        # note: deprecated since Python 3.8, replaced by ast.Constant
        self.add('"%s"' % node.s)


class UnparseJavascript(UnparsePython):
    """
    Unparse AST to generate Javascript script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def _ast_dict(self, node):
        """Add an AST Dict in output."""
        self.add(
            '{',
            self.make_list([[key, ': ', value]
                            for key, value in zip(node.keys, node.values)]),
            '}',
        )

    def _ast_functiondef(self, node):
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            'function %s(' % node.name,
            self.make_list(node.args.args),
            ') {',
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            '}',
        )

    def _ast_if(self, node):
        """Add an AST If in output."""
        self.add(
            self.fill,
            'if (',
            node.test,
            ') {',
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            '}',
        )
        if node.orelse:
            self.add(
                ' else {',
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                '}',
            )

    def _ast_pass(self, node):
        """Add an AST Pass in output."""
        pass


class UnparsePhp(UnparsePython):
    """
    Unparse AST to generate PHP script code.

    Note: only part of AST types are supported (just the types used by
    the script to test WeeChat scripting API).
    """

    __lineno__ = inspect.currentframe().f_lineno

    def _ast_assign(self, node):
        """Add an AST Assign in output."""
        self.add(
            self.fill,
            (self.prefix, '$'),
            [[target, ' = '] for target in node.targets],
            (self.prefix, None),
            node.value,
            ';',
        )

    def _ast_attribute(self, node):
        """Add an AST Attribute in output."""
        saved_prefix = self._prefix
        self._prefix = []
        if not node.attr.startswith('WEECHAT_'):
            self.add(node.value, '_')
        self.add(node.attr)
        self._prefix = saved_prefix

    def _ast_binop(self, node):
        """Add an AST BinOp in output."""
        if isinstance(node.op, ast.Add) and \
                (isinstance(node.left, (ast.Name, ast.Str)) or
                 isinstance(node.right, (ast.Name, ast.Str))):
            str_op = '.'
        else:
            str_op = self.binop[node.op.__class__.__name__]
        self.add(
            (self.prefix, '$'),
            node.left,
            ' %s ' % str_op,
            node.right,
            (self.prefix, None),
        )

    def _ast_call(self, node):
        """Add an AST Call in output."""
        self.add(
            node.func,
            '(',
            (self.prefix, '$'),
            self.make_list(node.args),
            (self.prefix, None),
            ')',
        )

    def _ast_constant(self, node):
        """Add an AST Constant in output."""
        if isinstance(node.s, str):
            self.add('"%s"' % node.s.replace('$', '\\$'))
        else:
            self.add(repr(node.s))

    def _ast_dict(self, node):
        """Add an AST Dict in output."""
        self.add(
            'array(',
            self.make_list([[key, ' => ', value]
                            for key, value in zip(node.keys, node.values)]),
            ')',
        )

    def _ast_expr(self, node):
        """Add an AST Expr in output."""
        if not isinstance(node.value, ast.Str):  # ignore docstrings
            self.add(
                self.fill,
                node.value,
                ';',
            )

    def _ast_functiondef(self, node):
        """Add an AST FunctionDef in output."""
        self.add(
            self.fill,
            self.fill,
            'function %s(' % node.name,
            (self.prefix, '$'),
            self.make_list(node.args.args),
            (self.prefix, None),
            ')',
            self.fill,
            '{',
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            '}',
        )

    def _ast_if(self, node):
        """Add an AST If in output."""
        self.add(
            self.fill,
            'if (',
            (self.prefix, '$'),
            node.test,
            (self.prefix, None),
            ')',
            self.fill,
            '{',
            self.indent,
            node.body,
            self.unindent,
            self.fill,
            '}',
        )
        if node.orelse:
            self.add(
                self.fill,
                'else',
                self.fill,
                '{',
                self.indent,
                node.orelse,
                self.unindent,
                self.fill,
                '}',
            )

    def _ast_pass(self, node):
        """Add an AST Pass in output."""
        pass

    def _ast_return(self, node):
        """Add an AST Return in output."""
        self.fill('return')
        if node.value:
            self.add(
                ' ',
                (self.prefix, '$'),
                node.value,
                (self.prefix, None),
                ';',
            )

    def _ast_str(self, node):
        """Add an AST Str in output."""
        # note: deprecated since Python 3.8, replaced by ast.Constant
        self.add('"%s"' % node.s.replace('$', '\\$'))


def get_languages():
    """Return a list of supported languages: ['python', 'perl', ...]."""

    members = [
        member
        for member in inspect.getmembers(sys.modules[__name__],
                                         predicate=inspect.isclass)
        if inspect.isclass(member[1]) and member[0].startswith('Unparse')
    ]

    languages = [
        name[7:].lower()
        for name, _ in sorted(members,
                              key=lambda member: member[1].__lineno__)
    ]

    return languages


LANGUAGES = get_languages()


def get_parser():
    """Get parser arguments."""
    all_languages = LANGUAGES + ['all']
    default_language = LANGUAGES[0]
    parser = argparse.ArgumentParser(
        description=('Unparse Python code from stdin and generate code in '
                     'another language (to stdout).\n\n'
                     'The code is read from stdin and generated code is '
                     'written on stdout.'))
    parser.add_argument(
        '-l', '--language',
        default=default_language,
        choices=all_languages,
        help='output language (default: %s)' % default_language)
    return parser


def get_stdin():
    """
    Return data from standard input.

    If there is nothing in stdin, wait for data until ctrl-D (EOF)
    is received.
    """
    data = ''
    inr = select.select([sys.stdin], [], [], 0)[0]
    if not inr:
        print('Enter the code to convert (Enter + ctrl+D to end)')
    while True:
        inr = select.select([sys.stdin], [], [], 0.1)[0]
        if not inr:
            continue
        new_data = os.read(sys.stdin.fileno(), 4096)
        if not new_data:  # EOF?
            break
        data += new_data.decode('utf-8')
    return data


def convert_to_language(code, language, prefix=''):
    """Convert Python code to a language."""
    class_name = 'Unparse%s' % language.capitalize()
    unparse_class = getattr(sys.modules[__name__], class_name)
    if prefix:
        print(prefix)
    output = StringIO()
    unparse_class(output=output).add(ast.parse(code))
    print(output.getvalue().lstrip())


def convert(code, language):
    """Convert Python code to one or all languages."""
    if language == 'all':
        for lang in LANGUAGES:
            convert_to_language(code, lang, '\n%s:' % lang)
    else:
        convert_to_language(code, language)


def main():
    """Main function."""
    parser = get_parser()
    args = parser.parse_args()

    code = get_stdin()
    if not code:
        print('ERROR: missing input')
        print()
        parser.print_help()
        sys.exit(1)

    convert(code, args.language)

    sys.exit(0)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020 Sébastien Helleu <flashcode@flashtux.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""
Scripts generator for WeeChat: build source of scripts in all languages to
test the scripting API.

This script can be run in WeeChat or as a standalone script
(during automatic tests, it is loaded as a WeeChat script).

It uses the following scripts:
- unparse.py: convert Python code to other languages (including Python itself)
- testapi.py: the WeeChat scripting API tests
"""

# pylint: disable=wrong-import-order,wrong-import-position

from __future__ import print_function
import argparse
import ast
from datetime import datetime
import inspect
try:
    from StringIO import StringIO  # python 2
except ImportError:
    from io import StringIO  # python 3
import os
import sys
import traceback

sys.dont_write_bytecode = True

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(SCRIPT_DIR)
from unparse import (
    UnparsePython,
    UnparsePerl,
    UnparseRuby,
    UnparseLua,
    UnparseTcl,
    UnparseGuile,
    UnparseJavascript,
    UnparsePhp,
)

RUNNING_IN_WEECHAT = True
try:
    import weechat
except ImportError:
    RUNNING_IN_WEECHAT = False


SCRIPT_NAME = 'testapigen'
SCRIPT_AUTHOR = 'Sébastien Helleu <flashcode@flashtux.org>'
SCRIPT_VERSION = '0.1'
SCRIPT_LICENSE = 'GPL3'
SCRIPT_DESC = 'Generate scripting API test scripts'

SCRIPT_COMMAND = 'testapigen'


class WeechatScript(object):  # pylint: disable=too-many-instance-attributes
    """
    A generic WeeChat script.

    This class must NOT be instanciated directly, use subclasses instead:
    PythonScript, PerlScript, ...
    """

    def __init__(self, unparse_class, tree, source_script, output_dir,
                 language, extension, comment_char='#',
                 weechat_module='weechat'):
        # pylint: disable=too-many-arguments
        self.unparse_class = unparse_class
        self.tree = tree
        self.source_script = os.path.realpath(source_script)
        self.output_dir = os.path.realpath(output_dir)
        self.language = language
        self.extension = extension
        self.script_name = 'testapi.%s' % extension
        self.script_path = os.path.join(self.output_dir, self.script_name)
        self.comment_char = comment_char
        self.weechat_module = weechat_module
        self.rename_functions()
        self.replace_variables()

    def comment(self, string):
        """Get a commented line."""
        return '%s %s' % (self.comment_char, string)

    def rename_functions(self):
        """Rename some API functions in the tree."""
        functions = {
            'prnt': 'print',
            'prnt_date_tags': 'print_date_tags',
            'prnt_y': 'print_y',
        }
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Call) and \
                    isinstance(node.func, ast.Attribute) and \
                    node.func.value.id == 'weechat':
                node.func.attr = functions.get(node.func.attr, node.func.attr)

    def replace_variables(self):
        """Replace script variables in string values."""
        variables = {
            'SCRIPT_SOURCE': self.source_script,
            'SCRIPT_NAME': self.script_name,
            'SCRIPT_PATH': self.script_path,
            'SCRIPT_AUTHOR': 'Sebastien Helleu',
            'SCRIPT_VERSION': '1.0',
            'SCRIPT_LICENSE': 'GPL3',
            'SCRIPT_DESCRIPTION': ('%s scripting API test' %
                                   self.language.capitalize()),
            'SCRIPT_LANGUAGE': self.language,
        }
        # count the total number of tests
        tests_count = 0
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Call) and \
                    isinstance(node.func, ast.Name) and \
                    node.func.id == 'check':
                tests_count += 1
        variables['SCRIPT_TESTS'] = str(tests_count)
        # replace variables
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Str) and \
                    node.s in variables:
                node.s = variables[node.s]

    def write_header(self, output):
        """Generate script header (just comments by default)."""
        comments = (
            '',
            '%s -- WeeChat %s scripting API testing' % (
                self.script_name, self.language.capitalize()),
            '',
            'WeeChat script automatically generated by testapigen.py.',
            'DO NOT EDIT BY HAND!',
            '',
            'Date: %s' % datetime.now(),
            '',
        )
        for line in comments:
            output.write(self.comment(line).rstrip() + '\n')

    def write(self):
        """Write script on disk."""
        print('Writing script %s... ' % self.script_path, end='')
        with open(self.script_path, 'w') as output:
            self.write_header(output)
            self.unparse_class(output).add(self.tree)
            output.write('\n')
            self.write_footer(output)
        print('OK')

    def write_footer(self, output):
        """Write footer (nothing by default)."""
        pass  # pylint: disable=unnecessary-pass


class WeechatPythonScript(WeechatScript):
    """A WeeChat script written in Python."""

    def __init__(self, tree, source_script, output_dir):
        super(WeechatPythonScript, self).__init__(
            UnparsePython, tree, source_script, output_dir, 'python', 'py')

    def rename_functions(self):
        # nothing to rename in Python
        pass

    def write_header(self, output):
        output.write('# -*- coding: utf-8 -*-\n')
        super(WeechatPythonScript, self).write_header(output)
        output.write('\n'
                     'import weechat')

    def write_footer(self, output):
        output.write('\n'
                     '\n'
                     'if __name__ == "__main__":\n'
                     '    weechat_init()\n')


class WeechatPerlScript(WeechatScript):
    """A WeeChat script written in Perl."""

    def __init__(self, tree, source_script, output_dir):
        super(WeechatPerlScript, self).__init__(
            UnparsePerl, tree, source_script, output_dir, 'perl', 'pl')

    def write_footer(self, output):
        output.write('\n'
                     'weechat_init();\n')


class WeechatRubyScript(WeechatScript):
    """A WeeChat script written in Ruby."""

    def __init__(self, tree, source_script, output_dir):
        super(WeechatRubyScript, self).__init__(
            UnparseRuby, tree, source_script, output_dir, 'ruby', 'rb')

    def rename_functions(self):
        super(WeechatRubyScript, self).rename_functions()
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Attribute) and \
                    node.value.id == 'weechat':
                node.value.id = 'Weechat'


class WeechatLuaScript(WeechatScript):
    """A WeeChat script written in Lua."""

    def __init__(self, tree, source_script, output_dir):
        super(WeechatLuaScript, self).__init__(
            UnparseLua, tree, source_script, output_dir, 'lua', 'lua',
            comment_char='--')

    def write_footer(self, output):
        output.write('\n'
                     'weechat_init()\n')


class WeechatTclScript(WeechatScript):
    """A WeeChat script written in Tcl."""

    def __init__(self, tree, source_script, output_dir):
        super(WeechatTclScript, self).__init__(
            UnparseTcl, tree, source_script, output_dir, 'tcl', 'tcl')

    def write_footer(self, output):
        output.write('\n'
                     'weechat_init\n')


class WeechatGuileScript(WeechatScript):
    """A WeeChat script written in Guile (Scheme)."""

    def __init__(self, tree, source_script, output_dir):
        super(WeechatGuileScript, self).__init__(
            UnparseGuile, tree, source_script, output_dir, 'guile', 'scm',
            comment_char=';')

    def write_footer(self, output):
        output.write('\n'
                     '(weechat_init)\n')


class WeechatJavascriptScript(WeechatScript):
    """A WeeChat script written in Javascript."""

    def __init__(self, tree, source_script, output_dir):
        super(WeechatJavascriptScript, self).__init__(
            UnparseJavascript, tree, source_script, output_dir,
            'javascript', 'js', comment_char='//')

    def write_footer(self, output):
        output.write('\n'
                     'weechat_init()\n')


class WeechatPhpScript(WeechatScript):
    """A WeeChat script written in PHP."""

    def __init__(self, tree, source_script, output_dir):
        super(WeechatPhpScript, self).__init__(
            UnparsePhp, tree, source_script, output_dir, 'php', 'php',
            comment_char='//')

    def write_header(self, output):
        output.write('<?php\n')
        super(WeechatPhpScript, self).write_header(output)

    def write_footer(self, output):
        output.write('\n'
                     'weechat_init();\n')

# ============================================================================


def update_nodes(tree):
    """
    Update the tests AST tree (in-place):
      1. add a print message in each test_* function
      2. add arguments in calls to check() function
    """
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef) and \
                node.name.startswith('test_'):
            # add a print at the beginning of each test function
            node.body.insert(
                0, ast.parse('weechat.prnt("", "  > %s");' % node.name))
        elif isinstance(node, ast.Call) and \
                isinstance(node.func, ast.Name) and \
                node.func.id == 'check':
            # add two arguments in the call to "check" function:
            #   1. the string representation of the test
            #   2. the line number in source (as string)
            # for example if this test is on line 50:
            #   check(weechat.test() == 123)
            # it becomes:
            #   check(weechat.test() == 123, 'weechat.test() == 123', '50')
            output = StringIO()
            unparsed = UnparsePython(output=output)
            unparsed.add(node.args[0])
            node.args.append(ast.Str(output.getvalue()))
            node.args.append(ast.Str(str(node.func.lineno)))


def get_tests(path):
    """Parse the source with tests and return the AST node."""
    test_script = open(path).read()
    tests = ast.parse(test_script)
    update_nodes(tests)
    return tests


def generate_scripts(source_script, output_dir):
    """Generate scripts in all languages to test the API."""
    ret_code = 0
    error = None
    try:
        for name, obj in inspect.getmembers(sys.modules[__name__]):
            if inspect.isclass(obj) and name != 'WeechatScript' and \
                    name.startswith('Weechat') and name.endswith('Script'):
                tests = get_tests(source_script)
                obj(tests, source_script, output_dir).write()
    except Exception as exc:  # pylint: disable=broad-except
        ret_code = 1
        error = 'ERROR: %s\n\n%s' % (str(exc), traceback.format_exc())
    return ret_code, error


def testapigen_cmd_cb(data, buf, args):
    """Callback for WeeChat command /testapigen."""

    def print_error(msg):
        """Print an error message on core buffer."""
        weechat.prnt('', '%s%s' % (weechat.prefix('error'), msg))

    try:
        source_script, output_dir = args.split()
    except ValueError:
        print_error('ERROR: invalid arguments for /testapigen')
        return weechat.WEECHAT_RC_OK
    if not weechat.mkdir_parents(output_dir, 0o755):
        print_error('ERROR: invalid directory: %s' % output_dir)
        return weechat.WEECHAT_RC_OK
    ret_code, error = generate_scripts(source_script, output_dir)
    if error:
        print_error(error)
    return weechat.WEECHAT_RC_OK if ret_code == 0 else weechat.WEECHAT_RC_ERROR


def get_parser_args():
    """Get parser arguments."""
    parser = argparse.ArgumentParser(
        description=('Generate WeeChat scripts in all languages '
                     'to test the API.'))
    parser.add_argument(
        'script',
        help='the path to Python script with tests')
    parser.add_argument(
        '-o', '--output-dir',
        default='.',
        help='output directory (defaults to current directory)')
    return parser.parse_args()


def main():
    """Main function (when script is not loaded in WeeChat)."""
    args = get_parser_args()
    ret_code, error = generate_scripts(args.script, args.output_dir)
    if error:
        print(error)
    sys.exit(ret_code)


if __name__ == '__main__':
    if RUNNING_IN_WEECHAT:
        weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION,
                         SCRIPT_LICENSE, SCRIPT_DESC, '', '')
        weechat.hook_command(
            SCRIPT_COMMAND,
            'Generate scripting API test scripts',
            'source_script output_dir',
            'source_script: path to source script (testapi.py)\n'
            '   output_dir: output directory for scripts',
            '',
            'testapigen_cmd_cb', '')
    else:
        main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020 Sébastien Helleu <flashcode@flashtux.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""
This script contains WeeChat scripting API tests
(it can not be run directly and can not be loaded in WeeChat).

It is parsed by testapigen.py, using Python AST (Abstract Syntax Trees),
to generate scripts in all supported languages (Python, Perl, Ruby, ...).
The resulting scripts can be loaded in WeeChat to test the scripting API.
"""

# pylint: disable=line-too-long,no-value-for-parameter

import weechat  # pylint: disable=import-error


def check(result, condition, lineno):
    """Display the result of a test."""
    if result:
        weechat.prnt('', '      TEST OK: ' + condition)
    else:
        weechat.prnt('',
                     'SCRIPT_SOURCE' + ':' + lineno + ':1: ' +
                     'ERROR: [' + 'SCRIPT_NAME' + '] condition is false: ' +
                     condition)


def test_plugins():
    """Test plugins functions."""
    check(weechat.plugin_get_name('') == 'core')
    check(weechat.plugin_get_name(weechat.buffer_get_pointer(weechat.buffer_search_main(), 'plugin')) == 'core')


def test_strings():
    """Test string functions."""
    check(weechat.charset_set('iso-8859-15') == 1)
    check(weechat.charset_set('') == 1)
    check(weechat.iconv_to_internal('iso-8859-15', 'abc') == 'abc')
    check(weechat.iconv_from_internal('iso-8859-15', 'abcd') == 'abcd')
    check(weechat.gettext('abcdef') == 'abcdef')
    check(weechat.ngettext('file', 'files', 1) == 'file')
    check(weechat.ngettext('file', 'files', 2) == 'files')
    check(weechat.strlen_screen('abcd') == 4)
    check(weechat.string_match('abcdef', 'abc*', 0) == 1)
    check(weechat.string_match('abcdef', 'abc*', 1) == 1)
    check(weechat.string_match('ABCDEF', 'abc*', 1) == 0)
    check(weechat.string_match_list('abcdef', '*,!abc*', 0) == 0)
    check(weechat.string_match_list('ABCDEF', '*,!abc*', 1) == 1)
    check(weechat.string_match_list('def', '*,!abc*', 0) == 1)
    check(weechat.string_eval_path_home('test ${abc}', {}, {'abc': '123'}, {}) == 'test 123')
    check(weechat.string_mask_to_regex('test*mask') == 'test.*mask')
    check(weechat.string_has_highlight('my test string', 'test,word2') == 1)
    check(weechat.string_has_highlight_regex('my test string', 'test|word2') == 1)
    check(weechat.string_format_size(0) == '0 bytes')
    check(weechat.string_format_size(1) == '1 byte')
    check(weechat.string_format_size(2097152) == '2.10 MB')
    check(weechat.string_format_size(420000000) == '420.00 MB')
    check(weechat.string_color_code_size('') == 0)
    check(weechat.string_color_code_size('test') == 0)
    str_color = weechat.color('yellow,red')
    check(weechat.string_color_code_size(str_color) == 7)
    check(weechat.string_remove_color('test', '?') == 'test')
    check(weechat.string_is_command_char('/test') == 1)
    check(weechat.string_is_command_char('test') == 0)
    check(weechat.string_input_for_buffer('test') == 'test')
    check(weechat.string_input_for_buffer('/test') == '')
    check(weechat.string_input_for_buffer('//test') == '/test')
    check(weechat.string_eval_expression("100 > 50", {}, {}, {"type": "condition"}) == '1')
    check(weechat.string_eval_expression("-50 < 100", {}, {}, {"type": "condition"}) == '1')
    check(weechat.string_eval_expression("18.2 > 5", {}, {}, {"type": "condition"}) == '1')
    check(weechat.string_eval_expression("0xA3 > 2", {}, {}, {"type": "condition"}) == '1')
    check(weechat.string_eval_expression("${buffer.full_name}", {}, {}, {}) == 'core.weechat')


def test_lists():
    """Test list functions."""
    ptr_list = weechat.list_new()
    check(ptr_list != '')
    check(weechat.list_size(ptr_list) == 0)
    item_def = weechat.list_add(ptr_list, 'def', weechat.WEECHAT_LIST_POS_SORT, '')
    check(weechat.list_size(ptr_list) == 1)
    item_abc = weechat.list_add(ptr_list, 'abc', weechat.WEECHAT_LIST_POS_SORT, '')
    check(weechat.list_size(ptr_list) == 2)
    check(weechat.list_search(ptr_list, 'abc') == item_abc)
    check(weechat.list_search(ptr_list, 'def') == item_def)
    check(weechat.list_search(ptr_list, 'ghi') == '')
    check(weechat.list_search_pos(ptr_list, 'abc') == 0)
    check(weechat.list_search_pos(ptr_list, 'def') == 1)
    check(weechat.list_search_pos(ptr_list, 'ghi') == -1)
    check(weechat.list_casesearch(ptr_list, 'abc') == item_abc)
    check(weechat.list_casesearch(ptr_list, 'def') == item_def)
    check(weechat.list_casesearch(ptr_list, 'ghi') == '')
    check(weechat.list_casesearch(ptr_list, 'ABC') == item_abc)
    check(weechat.list_casesearch(ptr_list, 'DEF') == item_def)
    check(weechat.list_casesearch(ptr_list, 'GHI') == '')
    check(weechat.list_casesearch_pos(ptr_list, 'abc') == 0)
    check(weechat.list_casesearch_pos(ptr_list, 'def') == 1)
    check(weechat.list_casesearch_pos(ptr_list, 'ghi') == -1)
    check(weechat.list_casesearch_pos(ptr_list, 'ABC') == 0)
    check(weechat.list_casesearch_pos(ptr_list, 'DEF') == 1)
    check(weechat.list_casesearch_pos(ptr_list, 'GHI') == -1)
    check(weechat.list_get(ptr_list, 0) == item_abc)
    check(weechat.list_get(ptr_list, 1) == item_def)
    check(weechat.list_get(ptr_list, 2) == '')
    weechat.list_set(item_def, 'def2')
    check(weechat.list_string(item_def) == 'def2')
    check(weechat.list_next(item_abc) == item_def)
    check(weechat.list_next(item_def) == '')
    check(weechat.list_prev(item_abc) == '')
    check(weechat.list_prev(item_def) == item_abc)
    weechat.list_remove(ptr_list, item_abc)
    check(weechat.list_size(ptr_list) == 1)
    check(weechat.list_get(ptr_list, 0) == item_def)
    check(weechat.list_get(ptr_list, 1) == '')
    weechat.list_remove_all(ptr_list)
    check(weechat.list_size(ptr_list) == 0)
    weechat.list_free(ptr_list)


def test_key():
    """Test key functions."""
    check(
        weechat.key_bind(
            'mouse',
            {
                '@chat(plugin.test):button1': 'hsignal:test_mouse',
                '@chat(plugin.test):wheelup': '/mycommand up',
                '@chat(plugin.test):wheeldown': '/mycommand down',
                '__quiet': '',
            }
        ) == 3)
    check(weechat.key_unbind('mouse', 'quiet:area:chat(plugin.test)') == 3)


def test_display():
    """Test display functions."""
    check(weechat.prefix('action') != '')
    check(weechat.prefix('error') != '')
    check(weechat.prefix('join') != '')
    check(weechat.prefix('network') != '')
    check(weechat.prefix('quit') != '')
    check(weechat.prefix('unknown') == '')
    check(weechat.color('green') != '')
    check(weechat.color('unknown') == '')


def completion_cb(data, completion_item, buf, completion):
    """Completion callback."""
    check(data == 'completion_data')
    check(completion_item == 'SCRIPT_NAME')
    check(weechat.completion_get_string(completion, 'args') == 'w')
    weechat.completion_list_add(completion, 'word_completed',
                                0, weechat.WEECHAT_LIST_POS_END)
    return weechat.WEECHAT_RC_OK


def command_cb(data, buf, args):
    """Command callback."""
    check(data == 'command_data')
    check(args == 'word_completed')
    return weechat.WEECHAT_RC_OK


def command_run_cb(data, buf, command):
    """Command_run callback."""
    check(data == 'command_run_data')
    check(command == '/cmd' + 'SCRIPT_NAME' + ' word_completed')
    return weechat.WEECHAT_RC_OK


def test_hooks():
    """Test function hook_command."""
    # hook_completion / hook_completion_args / and hook_command
    hook_cmplt = weechat.hook_completion('SCRIPT_NAME', 'description',
                                         'completion_cb', 'completion_data')
    hook_cmd = weechat.hook_command('cmd' + 'SCRIPT_NAME', 'description',
                                    'arguments', 'description arguments',
                                    '%(' + 'SCRIPT_NAME' + ')',
                                    'command_cb', 'command_data')
    weechat.command('', '/input insert /cmd' + 'SCRIPT_NAME' + ' w')
    weechat.command('', '/input complete_next')
    # hook_command_run
    hook_cmd_run = weechat.hook_command_run('/cmd' + 'SCRIPT_NAME' + '*',
                                            'command_run_cb', 'command_run_data')
    weechat.command('', '/input return')
    weechat.unhook(hook_cmd_run)
    weechat.unhook(hook_cmd)
    weechat.unhook(hook_cmplt)


def test_command():
    """Test command functions."""
    check(weechat.command('', '/mute') == 0)
    check(weechat.command_options('', '/mute', {'commands': '*,!print'}) == 0)
    check(weechat.command_options('', '/mute', {'commands': '*,!mute'}) == -1)


def infolist_cb(data, infolist_name, pointer, arguments):
    """Infolist callback."""
    infolist = weechat.infolist_new()
    check(infolist != '')
    item = weechat.infolist_new_item(infolist)
    check(item != '')
    check(weechat.infolist_new_var_integer(item, 'integer', 123) != '')
    check(weechat.infolist_new_var_string(item, 'string', 'test string') != '')
    check(weechat.infolist_new_var_pointer(item, 'pointer', '0xabcdef') != '')
    check(weechat.infolist_new_var_time(item, 'time', 1231231230) != '')
    return infolist


def test_infolist():
    """Test infolist functions."""
    hook_infolist = weechat.hook_infolist('infolist_test_script',
                                          'description', '', '',
                                          'infolist_cb', '')
    check(weechat.infolist_get('infolist_does_not_exist', '', '') == '')
    ptr_infolist = weechat.infolist_get('infolist_test_script', '', '')
    check(ptr_infolist != '')
    check(weechat.infolist_next(ptr_infolist) == 1)
    check(weechat.infolist_integer(ptr_infolist, 'integer') == 123)
    check(weechat.infolist_string(ptr_infolist, 'string') == 'test string')
    check(weechat.infolist_pointer(ptr_infolist, 'pointer') == '0xabcdef')
    check(weechat.infolist_time(ptr_infolist, 'time') == 1231231230)
    check(weechat.infolist_fields(ptr_infolist) == 'i:integer,s:string,p:pointer,t:time')
    check(weechat.infolist_next(ptr_infolist) == 0)
    weechat.infolist_free(ptr_infolist)
    weechat.unhook(hook_infolist)


def cmd_test_cb(data, buf, args):
    """Run all the tests."""
    weechat.prnt('', '>>>')
    weechat.prnt('', '>>> ------------------------------')
    weechat.prnt('', '>>> Testing ' + 'SCRIPT_LANGUAGE' + ' API')
    weechat.prnt('', '  > TESTS: ' + 'SCRIPT_TESTS')
    test_plugins()
    test_strings()
    test_lists()
    test_key()
    test_display()
    test_hooks()
    test_command()
    test_infolist()
    weechat.prnt('', '  > TESTS END')
    return weechat.WEECHAT_RC_OK


def weechat_init():
    """Main function."""
    weechat.register('SCRIPT_NAME', 'SCRIPT_AUTHOR', 'SCRIPT_VERSION',
                     'SCRIPT_LICENSE', 'SCRIPT_DESCRIPTION', '', '')
    weechat.hook_command('SCRIPT_NAME', '', '', '', '', 'cmd_test_cb', '')
